@@ -1,7 +1,7 @@
 #! /usr/bin/env python
 # IM - Infrastructure Manager
 # Copyright (C) 2011 - GRyCAP - Universitat Politecnica de Valencia
-# 
+#
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
 # the Free Software Foundation, either version 3 of the License, or
@@ -14,22 +14,32 @@
 #
 # You should have received a copy of the GNU General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
-
 """ Classes to encapsulate SSH operations using paramiko """
 
 import paramiko
+try:
+    import scp
+except:
+    pass
 import os
-import StringIO
+try:
+    from StringIO import StringIO
+except ImportError:
+    from io import StringIO
 from threading import Thread
+from stat import S_ISDIR
+
 
 class TimeOutException(Exception):
     """Timeout in the SSH execution"""
     pass
 
+
 class AuthenticationException(Exception):
     """Authentication error in the SSH connection"""
     pass
 
+
 class ThreadSSH(Thread):
     """Thread class to execute SSH with timeout"""
 
@@ -40,39 +50,47 @@ class ThreadSSH(Thread):
         self.command = None
         self.command_return = None
         self.client = None
-        
+
+    def close(self):
+        """
+        Close the SSH client connection
+        """
+        if self.client:
+            self.client.close()
+            self.client = None
+
     def run(self):
         if self.command:
             self.client = self.ssh.connect()
-            
+
+            channel = self.client.get_transport().open_session()
             if self.ssh.tty:
-                channel = self.client.get_transport().open_session()
                 channel.get_pty()
-                channel.exec_command(self.command + "\n")
-                stdout = channel.makefile()
-                stderr = channel.makefile_stderr()
-            else:
-                _, stdout, stderr = self.client.exec_command(self.command)
-            
+            channel.exec_command(self.command + "\n")
+            stdout = channel.makefile()
+            stderr = channel.makefile_stderr()
+            exit_status = channel.recv_exit_status()
+
             res_stdout = ""
             for line in stdout:
                 res_stdout += line
             res_stderr = ""
             for line in stderr:
                 res_stderr += line
-                
+
             if self.ssh.tty:
                 channel.close()
 
-            self.command_return = (res_stdout, res_stderr)
+            self.command_return = (res_stdout, res_stderr, exit_status)
+
 
 class SSH:
     """ Class to encapsulate SSH operations using paramiko """
 
-    def __init__(self, host, user, passwd, private_key = None, port=22):
+    def __init__(self, host, user, passwd, private_key=None, port=22):
         # Atributo para la version "thread"
         self.thread = None
-        
+
         self.tty = False
         self.port = port
         self.host = host
@@ -80,8 +98,8 @@ class SSH:
         self.password = passwd
         self.private_key = private_key
         self.private_key_obj = None
-        if (private_key != None and private_key.strip() != ""):
-            private_key_obj = StringIO.StringIO()
+        if (private_key is not None and private_key.strip() != ""):
+            private_key_obj = StringIO()
             if os.path.isfile(private_key):
                 pkfile = open(private_key)
                 for line in pkfile.readlines():
@@ -89,24 +107,26 @@ class SSH:
                 pkfile.close()
             else:
                 private_key_obj.write(private_key)
-                
+
             private_key_obj.seek(0)
-            self.private_key_obj = paramiko.RSAKey.from_private_key(private_key_obj)
+            self.private_key_obj = paramiko.RSAKey.from_private_key(
+                private_key_obj)
 
     def __str__(self):
-        res = "SSH: host: " + self.host + ", port: " + str(self.port) + ", user: " + self.username
-        if self.password != None:
+        res = "SSH: host: " + self.host + ", port: " + \
+            str(self.port) + ", user: " + self.username
+        if self.password is not None:
             res += ", password: " + self.password
-        if self.private_key != None:
+        if self.private_key is not None:
             res += ", private_key: " + self.private_key
         return res
 
-    def connect(self, time_out = None):
+    def connect(self, time_out=None):
         """ Establishes the connection with the SSH server
-        
+
             Arguments:
             - time_out: Timeout to connect.
-            
+
             Returns: a paramiko SSHClient connected with the server.
         """
         client = paramiko.SSHClient()
@@ -117,26 +137,26 @@ class SSH:
             # If both credentials are provided first try to use the password
             try:
                 client.connect(self.host, self.port, username=self.username,
-                           password=self.password, timeout=time_out)
+                               password=self.password, timeout=time_out)
             except paramiko.AuthenticationException:
                 # and then use the private key
                 client.connect(self.host, self.port, username=self.username,
-                           pkey = self.private_key_obj, timeout=time_out)
+                               pkey=self.private_key_obj, timeout=time_out)
         else:
             client.connect(self.host, self.port, username=self.username,
-                       password=self.password, timeout=time_out,
-                       pkey = self.private_key_obj)
-            
+                           password=self.password, timeout=time_out,
+                           pkey=self.private_key_obj)
+
         return client
 
-    def test_connectivity(self, time_out = None):
+    def test_connectivity(self, time_out=None):
         """ Tests if the SSH is active
-        
+
             Arguments:
             - time_out: Timeout to connect.
-            
+
             Returns: True if the connection is established or False otherwise
-            
+
             Raises:
                 Exception
         """
@@ -146,26 +166,26 @@ class SSH:
             return True
         except paramiko.AuthenticationException:
             raise AuthenticationException("Authentication Error!!")
-        except paramiko.SSHException, e:
+        except paramiko.SSHException as e:
             if str(e) == "No authentication methods available":
                 raise AuthenticationException("Authentication Error!!")
             return False
         except:
             return False
 
-    def execute(self, command, timeout = None):
+    def execute(self, command, timeout=None):
         """ Executes a command in the remote server
             The object must be connected.
-        
+
             Arguments:
             - command: The command to execute.
             - timeout: Timeout to connect.
-            
+
             Returns: A tuple (stdout, stderr, exit_code) with the output of the command and the exit code
         """
-        client = self.connect(time_out = timeout)
+        client = self.connect(time_out=timeout)
         channel = client.get_transport().open_session()
-        
+
         if self.tty:
             channel.get_pty()
 
@@ -173,7 +193,7 @@ class SSH:
         stdout = channel.makefile()
         stderr = channel.makefile_stderr()
         exit_status = channel.recv_exit_status()
-        
+
         res_stdout = ""
         for line in stdout:
             res_stdout += line
@@ -187,28 +207,43 @@ class SSH:
 
     def sftp_get(self, src, dest):
         """ Gets a file from the remote server
-        
+
             Arguments:
             - src: Source file in the remote server.
             - dest: Local destination path to copy.
         """
         client = self.connect()
         transport = client.get_transport()
-        sftp = paramiko.SFTPClient.from_transport(transport)
+
+        try:
+            sftp = paramiko.SFTPClient.from_transport(transport)
+            if not transport.active:
+                sftp = scp.SCPClient(transport)
+        except:
+            # in case of failure try to use scp
+            sftp = scp.SCPClient(transport)
+
         sftp.get(src, dest)
         sftp.close()
         transport.close()
 
     def sftp_get_files(self, src, dest):
         """ Gets a list of files from the remote server
-        
+
             Arguments:
             - src: A list with the source files in the remote server.
             - dest: A list with the local destination paths to copy.
         """
         client = self.connect()
         transport = client.get_transport()
-        sftp = paramiko.SFTPClient.from_transport(transport)
+        try:
+            sftp = paramiko.SFTPClient.from_transport(transport)
+            if not transport.active:
+                sftp = scp.SCPClient(transport)
+        except:
+            # in case of failure try to use scp
+            sftp = scp.SCPClient(transport)
+
         for file0, file1 in zip(src, dest):
             sftp.get(file0, file1)
         sftp.close()
@@ -216,13 +251,21 @@ class SSH:
 
     def sftp_put_files(self, files):
         """ Puts a list of files to the remote server
-        
+
             Arguments:
-            - files: A tuple where the first elements is the local source file to copy and the second element the destination paths in the remote server.
+            - files: A tuple where the first elements is the local source file to copy and the second
+                     element the destination paths in the remote server.
         """
         client = self.connect()
         transport = client.get_transport()
-        sftp = paramiko.SFTPClient.from_transport(transport)
+        try:
+            sftp = paramiko.SFTPClient.from_transport(transport)
+            if not transport.active:
+                sftp = scp.SCPClient(transport)
+        except:
+            # in case of failure try to use scp
+            sftp = scp.SCPClient(transport)
+
         for src, dest in files:
             sftp.put(src, dest)
         sftp.close()
@@ -230,51 +273,124 @@ class SSH:
 
     def sftp_put(self, src, dest):
         """ Puts a file to the remote server
-        
+
             Arguments:
             - src: Source local file to copy.
             - dest: Destination path in the remote server.
         """
         client = self.connect()
         transport = client.get_transport()
-        sftp = paramiko.SFTPClient.from_transport(transport)
+        try:
+            sftp = paramiko.SFTPClient.from_transport(transport)
+            if not transport.active:
+                sftp = scp.SCPClient(transport)
+        except:
+            # in case of failure try to use scp
+            sftp = scp.SCPClient(transport)
         sftp.put(src, dest)
         sftp.close()
         transport.close()
 
+    def sftp_get_dir(self, src, dest):
+        """ Gets recursively a directory from the remote server
+
+            Arguments:
+            - src: Source directory in the remote server to copy.
+            - dest: Local destination path.
+        """
+        client = self.connect()
+        transport = client.get_transport()
+        sftp = paramiko.SFTPClient.from_transport(transport)
+
+        files = self.sftp_walk(src, None, sftp)
+
+        for filename in files:
+            dirname = os.path.dirname(filename)
+            if not os.path.exists(dirname):
+                os.mkdir(dirname)
+            full_dest = filename.replace(src, dest)
+            sftp.get(filename, full_dest)
+
+        sftp.close()
+        transport.close()
+
+    def sftp_walk(self, src, files=None, sftp=None):
+        """ Gets recursively the list of items in a directory from the remote server
+
+            Arguments:
+            - src: Source directory in the remote server to copy.
+        """
+        close = False
+        if not sftp:
+            client = self.connect()
+            transport = client.get_transport()
+            sftp = paramiko.SFTPClient.from_transport(transport)
+            close = True
+
+        folders = []
+        if not files:
+            files = []
+        for f in sftp.listdir_attr(src):
+            if S_ISDIR(f.st_mode):
+                folder = os.path.join(src, f.filename)
+                folders.append(folder)
+            else:
+                filename = os.path.join(src, f.filename)
+                files.append(filename)
+
+        for folder in folders:
+            self.sftp_walk(folder, files, sftp)
+
+        if close:
+            sftp.close()
+            transport.close()
+
+        return files
+
     def sftp_put_dir(self, src, dest):
         """ Puts recursively the contents of a directory to the remote server
-        
+
             Arguments:
             - src: Source local directory to copy.
             - dest: Destination path in the remote server.
         """
         if os.path.isdir(src):
+            if src.endswith("/"):
+                src = src[:-1]
             client = self.connect()
             transport = client.get_transport()
-            sftp = paramiko.SFTPClient.from_transport(transport)
-        
+            try:
+                sftp = paramiko.SFTPClient.from_transport(transport)
+                sftp_avail = transport.active
+            except:
+                # in case of failure try to use scp
+                sftp = scp.SCPClient(transport)
+                sftp_avail = False
+
             for dirname, dirnames, filenames in os.walk(src):
                 for subdirname in dirnames:
                     src_path = os.path.join(dirname, subdirname)
-                    dest_path = os.path.join(dest, src_path[len(src)+1:])
-                    try:
-                        # si existe no lo intentamos crear
-                        sftp.stat(dest_path)
-                    except:
-                        sftp.mkdir(dest_path)
+                    dest_path = os.path.join(dest, src_path[len(src) + 1:])
+                    if sftp_avail:
+                        try:
+                            # if it exists we do not try to create it
+                            sftp.stat(dest_path)
+                        except:
+                            sftp.mkdir(dest_path)
+                    else:
+                        self.execute("mkdir -p %s" % dest_path)
                 for filename in filenames:
                     src_file = os.path.join(dirname, filename)
-                    dest_file = os.path.join(dest, dirname[len(src)+1:],
+                    dest_file = os.path.join(dest, dirname[len(src) + 1:],
                                              filename)
                     sftp.put(src_file, dest_file)
-                    
+
             sftp.close()
             transport.close()
 
     def sftp_put_content(self, content, dest):
         """ Puts the contents of a string in a remote file
-        
+
             Arguments:
             - content: The string to put into the remote file.
             - dest: Destination path in the remote server.
@@ -288,35 +404,46 @@ class SSH:
         sftp.close()
         transport.close()
 
-    def sftp_mkdir(self, directory):
+    def sftp_mkdir(self, directory, mode=0o777):
         """ Creates a remote directory
-        
+
             Arguments:
             - directory: Name of the directory in the remote server.
-            
+
             Returns: True if the directory is created or False if it exists.
         """
-        client = self.connect()
-        transport = client.get_transport()
-        sftp = paramiko.SFTPClient.from_transport(transport)
         try:
-            # si existe no lo intentamos crear
-            sftp.stat(directory)
-            res = False
+            client = self.connect()
+            transport = client.get_transport()
+            sftp = paramiko.SFTPClient.from_transport(transport)
+            sftp_avail = transport.active
         except:
-            sftp.mkdir(directory)
-            res = True
-            
-        sftp.close()
-        transport.close()
+            sftp_avail = False
+
+        if sftp_avail:
+            try:
+                # if it exists we do not try to create it
+                sftp.stat(directory)
+                res = False
+            except:
+                sftp.mkdir(directory, mode)
+                res = True
+
+            sftp.close()
+            transport.close()
+        else:
+            # use mkdir over ssh to create the directory
+            _, _, status = self.execute("mkdir -p %s" % directory)
+            res = status == 0
+
         return res
 
     def sftp_list(self, directory):
         """ List the contents of a remote directory
-        
+
             Arguments:
             - directory: Name of the directory in the remote server.
-            
+
             Returns: A list with the contents of the directory
                      (see paramiko.SFTPClient.listdir)
         """
@@ -331,10 +458,10 @@ class SSH:
     def sftp_list_attr(self, directory):
         """ Return a list containing SFTPAttributes objects corresponding to
             files in the given path.
-        
+
             Arguments:
             - directory: Name of the directory in the remote server.
-            
+
             Returns: A list containing SFTPAttributes object
                      (see paramiko.SFTPClient.listdir_attr)
         """
@@ -348,20 +475,30 @@ class SSH:
 
     def getcwd(self):
         """ Get the current working directory.
-            
+
             Returns: The current working directory.
         """
-        client = self.connect()
-        transport = client.get_transport()
-        sftp = paramiko.SFTPClient.from_transport(transport)
-        cwd = sftp.getcwd()
-        sftp.close()
-        transport.close()
-        return cwd
-    
-    def execute_timeout(self, command, timeout, retry = 1, kill_command = None):
+        try:
+            client = self.connect()
+            transport = client.get_transport()
+            sftp = paramiko.SFTPClient.from_transport(transport)
+            sftp_avail = transport.active
+        except:
+            sftp_avail = False
+
+        if sftp_avail:
+            cwd = sftp.getcwd()
+            sftp.close()
+            transport.close()
+        else:
+            # use rm over ssh to delete the file
+            cwd, _, _ = self.execute("pwd")
+
+        return str(cwd.strip("\n"))
+
+    def execute_timeout(self, command, timeout, retry=1, kill_command=None):
         """ Executes a command waiting for a timeout, and send a kill comand
-        
+
             Arguments:
             - command: Command to execute.
             - timeout: Timeout to wait for.
@@ -370,7 +507,7 @@ class SSH:
             - kill_command: A command sent when the timeout is expired,
                 to clean the environment.
                 (Optional, default None)
-            
+
             Returns: A tuple (stdout, stderr) with the output of the command
         """
         cont = 0
@@ -382,15 +519,72 @@ class SSH:
             self.thread.join(timeout)
 
             if self.thread.isAlive():
-                self.thread.client.close()
+                self.thread.close()
                 self.thread.join(2)
                 if kill_command:
                     self.execute(kill_command)
                 self.thread = None
             else:
                 res = self.thread.command_return
-                self.thread.client.close()
+                self.thread.close()
                 self.thread = None
                 return res
-            
+
         raise TimeOutException("Error: Timeout")
+
+    def sftp_remove(self, path):
+        """ Delete a file, if possible.
+
+            Arguments:
+            - path: Name of the file in the remote server to delete.
+
+            Returns: True if the file is deleted or False if it exists.
+        """
+        try:
+            client = self.connect()
+            transport = client.get_transport()
+            sftp = paramiko.SFTPClient.from_transport(transport)
+            sftp_avail = transport.active
+        except:
+            sftp_avail = False
+
+        if sftp_avail:
+            res = sftp.remove(path)
+            sftp.close()
+            transport.close()
+        else:
+            # use rm over ssh to delete the file
+            _, _, status = self.execute("rm -f %s" % path)
+            res = status == 0
+
+        return res
+
+    def sftp_chmod(self, path, mode):
+        """
+        Change the mode (permissions) of a file.  The permissions are
+        unix-style and identical to those used by python's C{os.chmod}
+        function.
+
+            Arguments:
+            - path: String with the path of the file to change the permissions of
+            - mode: Int with the new permissions
+        """
+        try:
+            client = self.connect()
+            transport = client.get_transport()
+            sftp = paramiko.SFTPClient.from_transport(transport)
+            sftp_avail = transport.active
+        except:
+            sftp_avail = False
+
+        if sftp_avail:
+            sftp.chmod(path, mode)
+            res = True
+            sftp.close()
+            transport.close()
+        else:
+            # use chmod over ssh to change permissions
+            _, _, status = self.execute("chmod %s %s" % (oct(mode), path))
+            res = status == 0
+
+        return res
