@@ -15,7 +15,9 @@
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
 
-__all__ = ['auth','bottle','CloudManager','config','ConfManager','db','ganglia','HTTPHeaderTransport','ImageManager','InfrastructureInfo','InfrastructureManager','parsetab','radl','recipe','request','REST','SSH','timedcall','uriparse','VMRC','xmlobject']
-__version__ = '1.1.1'
+__all__ = ['auth', 'CloudInfo', 'config', 'ConfManager', 'db',
+           'InfrastructureInfo', 'InfrastructureManager', 'recipe', 'request', 'REST', 'retry',
+           'ServiceRequests', 'SSH', 'SSHRetry', 'timedcall', 'UnixHTTPAdapter', 'uriparse',
+           'VirtualMachine', 'VMRC', 'xmlobject']
+__version__ = '1.7.6'
 __author__ = 'Miguel Caballer'
-
